cEmitter = gideros.class(Sprite)
 
function cEmitter:init(p)
	self.active = false
	self.t = cEmitter:deepcopy(p)
	self.template = cEmitter:deepcopy(p)
	self.maxParticles = self.t["maxParticles"]
	self.life = self.t["particleLifeSpan"]
	self.particles = {}
	self.emissionRate = self.maxParticles/self.life
	
	for i=1,self.maxParticles,1 do
		self.particles[i] = cEmitter:deepcopy(cEmitter:createParticle(self.t))
	end

	
end

function cEmitter:createParticle(t)
	local p = {}
	local bitmap = Bitmap.new(Texture.new(t["textureName"]))
	local bw = bitmap:getWidth()
	local bh = bitmap:getHeight()
	bitmap:setPosition(-bw/2,-bh/2)
	p.sprite = Sprite.new()
	p.sprite:addChild(bitmap)
	p.sprite:setVisible(false)
	stage:addChild(p.sprite)
	p.active = false
	
	local life = t["particleLifeSpan"]
	local lifeVar = t["particleLifespanVariance"]
	local sourcePosition = t["sourcePosition"]
	local posVar = t["sourcePositionVariance"]
	local startSize = t["startParticleSize"]
	local startSizeVar = t["startParticleSizeVariance"]
	local endSize = t["finishParticleSize"]
	local endSizeVar = t["FinishParticleSizeVariance"]
	local startSpin = t["rotationStart"]
	local startSpinVar = t["rotationStartVariance"]
	local endSpin = t["rotationEnd"]
	local endSpinVar = t["rotationEndVariance"]
	local angle = t["angle"]
	local angleVar = t["angleVariance"]
	


	local start = {
		r = clampf(t["startColor"].red + t["startColorVariance"].red * CCRANDOM_MINUS1_1(),0,1),
		g = clampf(t["startColor"].green + t["startColorVariance"].green * CCRANDOM_MINUS1_1(),0,1),
		b = clampf(t["startColor"].blue + t["startColorVariance"].blue * CCRANDOM_MINUS1_1(),0,1),
		a = clampf(t["startColor"].alpha + t["startColorVariance"].alpha * CCRANDOM_MINUS1_1(),0,1)
	}
	local End = {
		r = clampf(t["finishColor"].red + t["finishColorVariance"].red * CCRANDOM_MINUS1_1(),0,1),
		g = clampf(t["finishColor"].green + t["finishColorVariance"].green * CCRANDOM_MINUS1_1(),0,1),
		b = clampf(t["finishColor"].blue + t["finishColorVariance"].blue * CCRANDOM_MINUS1_1(),0,1),
		a = clampf(t["finishColor"].alpha + t["finishColorVariance"].alpha * CCRANDOM_MINUS1_1(),0,1)
	}

 	p.timeToLive = life * lifeVar * CCRANDOM_MINUS1_1()
	p.timeToLive = MAX(0,p.timeToLive)
	p.timeToLive = 1
	p.pos = {
		x = sourcePosition.x + posVar.x * CCRANDOM_MINUS1_1(),
		y = sourcePosition.y + posVar.y * CCRANDOM_MINUS1_1()
	}
	p.color = start
	p.deltaColor = {
		r = (End.r - start.r) / p.timeToLive,
		g = (End.g - start.g) / p.timeToLive,
		b = (End.b - start.b) / p.timeToLive,
		a = (End.a - start.a) / p.timeToLive
	}
	local startS = startSize + startSizeVar * CCRANDOM_MINUS1_1()
	startS = MAX(0,startS)
	
	
	p.size = startS
	p.deltaSize = 0
	if endSize == -1 then 
		p.deltaSize = 0
	else
		local endS = endSize + endSizeVar * CCRANDOM_MINUS1_1()
		endS = MAX(0,endS)
		p.deltaSize = (endS - startS) / p.timeToLive
	end

	local startA = startSpin + startSpinVar * CCRANDOM_MINUS1_1()
	local endA = endSpin + endSpinVar * CCRANDOM_MINUS1_1()
	p.rotation = startA
	p.deltaRotation = (endA - startA) / p.timeToLive

	local w = p.sprite:getWidth()
	local h = p.sprite:getHeight()
	p.startPos = ccpMult({x=p.sprite:getX()-w/2,y=p.sprite:getY()-h/2},1)
	
	local a = CC_DEGREES_TO_RADIANS( angle + angleVar * CCRANDOM_MINUS1_1() )
	p.pType = t["emitterType"]
	local v = { x=math.cos(a),y=math.sin(a)}
	local s = t["speed"] + t["speedVariance"] * CCRANDOM_MINUS1_1()
	local startRadius = t["maxRadius"]+t["maxRadiusVariance"]*CCRANDOM_MINUS1_1()
	local endRadius = t["endRadius"]+t["endRadiusVariance"]*CCRANDOM_MINUS1_1()
	local er = 0
	if endRadius == -1 then er = 0 else er = (endRadius-startRadius)/p.timeToLive end
	p.mode = {
		A = {
			dir = ccpMult(v,s),
			radialAccel = t["radialAcceleration"]+t["radialAccelVariance"]*CCRANDOM_MINUS1_1(),
			tangentialAccel = t["tangentialAcceleration"]+t["tangentialAccelVariance"]*CCRANDOM_MINUS1_1(),
			gravity = t["gravity"]
		},
		B = {
			angle = a,
			degreesPerSecond = CC_DEGREES_TO_RADIANS(t["rotatePerSecond"]+t["rotatePerSecondVariance"]*CCRANDOM_MINUS1_1()),
			radius = startRadius,
			deltaRadius = er
		}
	}
	return p
end

function cEmitter:addParticle(x,y)
	for i=1,self.maxParticles,1 do
		if self.particles[i].active == false then
			cEmitter:reset(self.particles[i],self.template,x,y)
			break
		end
	end
end

function cEmitter:reset(p,pSettings,sx,sy)
	t = emitter:deepcopy(pSettings)
	p.sprite:setVisible(true)
	local life = t["particleLifeSpan"]
	local lifeVar = t["particleLifespanVariance"]
	local sourcePosition = {}
	if sx ~= nil and sy ~=nil then
		sourcePosition = {x=sx,y=sy}
	else
		sourcePosition = t["sourcePosition"]
	end
	local posVar = t["sourcePositionVariance"]
	local startSize = t["startParticleSize"]
	local startSizeVar = t["startParticleSizeVariance"]
	local endSize = t["finishParticleSize"]
	local endSizeVar = t["FinishParticleSizeVariance"]
	local startSpin = t["rotationStart"]
	local startSpinVar = t["rotationStartVariance"]
	local endSpin = t["rotationEnd"]
	local endSpinVar = t["rotationEndVariance"]
	local angle = t["angle"]
	local angleVar = t["angleVariance"]

	local start = {
		r = clampf(t["startColor"].red + t["startColorVariance"].red * CCRANDOM_MINUS1_1(),0,1),
		g = clampf(t["startColor"].green + t["startColorVariance"].green * CCRANDOM_MINUS1_1(),0,1),
		b = clampf(t["startColor"].blue + t["startColorVariance"].blue * CCRANDOM_MINUS1_1(),0,1),
		a = clampf(t["startColor"].alpha + t["startColorVariance"].alpha * CCRANDOM_MINUS1_1(),0,1)
	}
	local End = {
		r = clampf(t["finishColor"].red + t["finishColorVariance"].red * CCRANDOM_MINUS1_1(),0,1),
		g = clampf(t["finishColor"].green + t["finishColorVariance"].green * CCRANDOM_MINUS1_1(),0,1),
		b = clampf(t["finishColor"].blue + t["finishColorVariance"].blue * CCRANDOM_MINUS1_1(),0,1),
		a = clampf(t["finishColor"].alpha + t["finishColorVariance"].alpha * CCRANDOM_MINUS1_1(),0,1)
	}

 	p.timeToLive = life * lifeVar * CCRANDOM_MINUS1_1()
	p.timeToLive = MAX(0,p.timeToLive)
	p.timeToLive = 1
	p.pos = {
		x = sourcePosition.x + posVar.x * CCRANDOM_MINUS1_1(),
		y = sourcePosition.y + posVar.y * CCRANDOM_MINUS1_1()
	}
	p.color = start
	p.deltaColor = {
		r = (End.r - start.r) / p.timeToLive,
		g = (End.g - start.g) / p.timeToLive,
		b = (End.b - start.b) / p.timeToLive,
		a = (End.a - start.a) / p.timeToLive
	}
	local startS = startSize + startSizeVar * CCRANDOM_MINUS1_1()
	startS = MAX(0,startS)
	
	
	p.size = startS
	p.deltaSize = 0
	if endSize == -1 then 
		p.deltaSize = 0
	else
		local endS = endSize + endSizeVar * CCRANDOM_MINUS1_1()
		endS = MAX(0,endS)
		p.deltaSize = (endS - startS) / p.timeToLive
	end

	local startA = startSpin + startSpinVar * CCRANDOM_MINUS1_1()
	local endA = endSpin + endSpinVar * CCRANDOM_MINUS1_1()
	p.rotation = startA
	p.deltaRotation = (endA - startA) / p.timeToLive

	local w = p.sprite:getWidth()
	local h = p.sprite:getHeight()
	p.startPos = ccpMult({x=p.sprite:getX()-w/2,y=p.sprite:getY()-h/2},1)
	
	local a = CC_DEGREES_TO_RADIANS( angle + angleVar * CCRANDOM_MINUS1_1() )
	p.pType = t["emitterType"]
	local v = { x=math.cos(a),y=math.sin(a)}
	local s = t["speed"] + t["speedVariance"] * CCRANDOM_MINUS1_1()
	local startRadius = t["maxRadius"]+t["maxRadiusVariance"]*CCRANDOM_MINUS1_1()
	local endRadius = t["endRadius"]+t["endRadiusVariance"]*CCRANDOM_MINUS1_1()
	local er = 0
	if endRadius == -1 then er = 0 else er = (endRadius-startRadius)/p.timeToLive end
	p.mode = {
		A = {
			dir = ccpMult(v,s),
			radialAccel = t["radialAcceleration"]+t["radialAccelVariance"]*CCRANDOM_MINUS1_1(),
			tangentialAccel = t["tangentialAcceleration"]+t["tangentialAccelVariance"]*CCRANDOM_MINUS1_1(),
			gravity = t["gravity"]
		},
		B = {
			angle = a,
			degreesPerSecond = CC_DEGREES_TO_RADIANS(t["rotatePerSecond"]+t["rotatePerSecondVariance"]*CCRANDOM_MINUS1_1()),
			radius = startRadius,
			deltaRadius = er
		}
	}
	p.active = true
	
end

function cEmitter:stop()
	self.active = false
end

function cEmitter:start()
	self.active = true
end

function cEmitter:update(dt,p)
	if p.active == false then return end
	--if not self.disabled then
	
	p.timeToLive = p.timeToLive - dt
	if p.timeToLive < 0 then
		p.active = false
		p.sprite:setVisible(false)
		return
	end
	if p.pType == 1 then
		local tmp = {x=0,y=0}
		local radial = {x=0,y=0}
		local tangential = {x=0,y=0}
		if p.pos.x or p.pos.y then
			radial = ccpNormalize(p.pos)
		end
		tangential = radial
		radial = ccpMult(radial,p.mode.A.radialAccel)
		local newy = tangential.x
		tangential.x = -tangential.y
		tangential.y = newy
		tangential = ccpMult(tangential,p.mode.A.tangentialAccel)
		tmp = ccpAdd(ccpAdd(radial,tangential),p.mode.A.gravity)
		tmp = ccpMult(tmp,dt)
		p.mode.A.dir = ccpAdd(p.mode.A.dir,tmp)
		tmp = ccpMult(p.mode.A.dir,dt)
		p.pos = ccpAdd(p.pos,tmp)
	end
	
	if p.pType == 2 then
	
	end
	
	--color
	p.color.r =  p.color.r + ( p.deltaColor.r * dt )
	p.color.g =  p.color.g + ( p.deltaColor.g * dt )
	p.color.b =  p.color.b + ( p.deltaColor.b * dt )
	p.color.a =  p.color.a + ( p.deltaColor.a * dt )
	p.sprite:setColorTransform(p.color.r,p.color.g,p.color.b,p.color.a)
	
	--size
	p.size = p.size + (p.deltaSize * dt)
	p.size = MAX( 0, p.size)
	p.sprite:setScale(p.size,p.size)
	

	
	--local currentPosition = {x=self.sprite:getX(),y=self.sprite:getY()}
	local newpos = {x=0,y=0}
	local diff = ccpSub({x=0,y=0},p.startPos)
	newpos = ccpSub(p.pos,diff)
	newpos = p.pos
	
	local w = p.sprite:getWidth()
	local h = p.sprite:getHeight()
	p.sprite:setPosition(newpos.x,newpos.y)
	
	--angle
	p.rotation = p.rotation + (p.deltaRotation * dt)

	p.sprite:setRotation(p.rotation)
	
	--self.sprite:setRotation(self.rotation)


	--end

end

function cEmitter:deepcopy(object)
    local lookup_table = {}
    local function _copy(object)
        if type(object) ~= "table" then
            return object
        elseif lookup_table[object] then
            return lookup_table[object]
        end
        local new_table = {}
        lookup_table[object] = new_table
        for index, value in pairs(object) do
            new_table[_copy(index)] = _copy(value)
        end
        return setmetatable(new_table, getmetatable(object))
    end
    return _copy(object)
end